import os
from typing import Optional

import requests

from kaginawa.exceptions import KaginawaError
from kaginawa.models import (
    KaginawaEnrichResponse,
    KaginawaFastGPTResponse,
    KaginawaSummarizationResponse,
)


class Kaginawa:
    """The main client to the Kagi API"""

    def __init__(
        self,
        token: Optional[str] = None,
        session: Optional[requests.Session] = None,
        api_base: str = "https://kagi.com/api",
    ):
        """Create a new instance of the Kagi API wrapper.

        Parameters:
          token (Optional[str], optional): The API access token to authenticate
            requests. If this value is unset the library will attempt to read the
            environment variable KAGI_API_KEY. If both are unset an exception will
            be raised.

          session (Optional[requests.Session], optional): An optional `requests`
            session object to use for sending HTTP requests. Defaults to `None`.

          api_base (str, optional): The base URL for the Kagi API endpoint.
            Defaults to "https://kagi.com/api".
        """

        try:
            token = token or os.environ["KAGI_API_KEY"]
        except KeyError as e:
            raise KaginawaError("Value for `token` not given and env KAGI_API_KEY is unset") from e

        self.token = token
        self.api_base = api_base

        if not session:
            session = requests.Session()

        self.session = session

        self.session.headers = {"Authorization": f"Bot {self.token}"}

    def generate(self, query: str, cache: Optional[bool] = None):
        """Generate a FastGPT response from a text query.

        Parameters:
          query (str): The prompt to send to FastGPT.

          cache (Optional[bool], optional): Allow cached responses. Defaults to `None`.

        Returns:
          KaginawaFastGPTResponse: A model representing the response.
        """
        try:
            optional_params = {}

            if cache is not None:
                optional_params["cache"] = cache

            res = self.session.post(
                f"{self.api_base}/v0/fastgpt",
                json={
                    "query": query,
                    **optional_params,
                },
            )

            res.raise_for_status()

            raw_response = res.json()
        except requests.RequestException as e:
            raise KaginawaError("Error calling /v0/fastgpt") from e

        return KaginawaFastGPTResponse.from_raw(raw_response)

    def enrich_web(self, query: str):
        """Query the Teclis index for relevant web results for a given query.

        Parameters:
          query (str): The search query.

        Returns:
          KaginawaEnrichWebResponse: A model representing the response.
        """

        try:
            res = self.session.get(
                f"{self.api_base}/v0/enrich/web",
                params={"q": query},
            )
            res.raise_for_status()

            raw_response = res.json()
        except requests.RequestException as e:
            raise KaginawaError("Error calling /v0/enrich/web") from e

        return KaginawaEnrichResponse.from_raw(raw_response).results[0]

    def enrich_news(self, query: str):
        """Query the Teclis index for relevant web results for a given query.

        Parameters:
          query (str): The search query.

        Returns:
          KaginawaEnrichWebResponse: A model representing the response.
        """

        try:
            res = self.session.get(
                f"{self.api_base}/v0/enrich/news",
                params={"q": query},
            )
            res.raise_for_status()

            raw_response = res.json()
        except requests.RequestException as e:
            raise KaginawaError("Error calling /v0/enrich/web") from e

        return KaginawaEnrichResponse.from_raw(raw_response).results[0]

    def summarize(
        self,
        url: Optional[str] = None,
        text: Optional[str] = None,
        engine: Optional[str] = None,
        summary_type: Optional[str] = None,
        target_language: Optional[str] = None,
        cache: Optional[bool] = None,
    ):
        """Summarize a URL or text snippet.

        Parameters:
          url (Optional[str], optional): The URL to summarize. This option is exclusive
            with the `text` parameter. Defaults to `None`.

          text (Optional[str], optional): The text snippet to summarize. This option is
            exclusive with the `url` parameter. Defaults to `None`.

          engine (Optional[str], optional): The summarization engine to use. There is a
            helper enum `KaginawaSummarizationEngine` with the valid values for this
            parameter. Defaults to `None`.

          summary_type (Optional[str], optional): The 'kind' of summary you would like
            the model to use. There is a helper enum KaginawaSummaryType with the valid
            values for this parameter. Defaults to `None`.

          target_language: (Optional[str], optional): The language CODE (eg. EN, ZH, FR)
            corresponding to the language you would like the summary in. See
            https://help.kagi.com/kagi/api/summarizer.html for valid language codes.
            Defaults to `None`.

          cache (Optional[bool], optional): Allow cached responses. Defaults to `None`.

        Returns:
          KaginawaSummarizationResponse: A model representing the response.
        """
        try:
            params = {}

            if not (bool(url) ^ bool(text)):
                raise KaginawaError("You must provide exactly one of 'url' or 'text'.")

            if url:
                params["url"] = url

            if text:
                params["text"] = text

            if engine:
                params["engine"] = engine

            if summary_type:
                params["summary_type"] = summary_type

            if target_language:
                params["target_language"] = target_language

            if cache is not None:
                params["cache"] = cache

            res = self.session.post(
                f"{self.api_base}/v0/summarize",
                data=params,
            )

            raw_response = res.json()
        except requests.RequestException as e:
            raise KaginawaError("Error calling /v0/summarize") from e

        return KaginawaSummarizationResponse.from_raw(raw_response)



from dataclasses import dataclass
from datetime import datetime, timedelta
from enum import StrEnum
from typing import Any, Dict, List


@dataclass
class KaginawaReference:
    title: str
    """The <title> of the reference."""

    snippet: str
    """A short blurb of text describing the reference."""

    url: str
    """The URL of the reference."""


@dataclass
class KaginawaResponse:
    id: str
    """The unique id of the API response."""

    node: str
    """The 'node' which is really a region the API response came from."""

    duration: timedelta
    """How long the request took to process (always 0 as of 2023-11-05)."""

    api_balance: float
    """How much money (in dollars) you have left on your account."""

    @classmethod
    def from_raw(cls, raw_response: Dict[str, Any], **kwargs):
        """Create a model from the raw API response.

        It's not expected that end-users of this library will ever need to call this
        method directly. It is called for you when invoking methods on the client.

        This method uses 'cooperative' inheritance due to the nature of external
        constructors. Subclasses should call `super().from_raw` with all of the
        subclass specific parameters passed in `**kwargs`.

        Parameters:
          raw_response (Dict[str, Any]): The raw response from the Kagi API.

          **kwargs: Other parameters to pass to the class constructor.

        Returns:
          KaginawaResponse: A new instance of the response object.
        """

        duration = timedelta(milliseconds=raw_response["meta"]["ms"])
        raw_response["meta"].pop("ms")

        return cls(
            duration=duration,
            **raw_response["meta"],
            **kwargs,
        )


@dataclass
class KaginawaFastGPTResponse(KaginawaResponse):
    output: str
    """The output of FastGPT."""

    tokens: int
    """The number of the tokens in the response (always 0 as of 2023-11-05)."""

    references: List[KaginawaReference]
    """A list of web pages that were provided as context to the LLM."""

    @classmethod
    def from_raw(cls, raw_response, **kwargs):
        """Create a model from the raw API response.

        It's not expected that end-users of this library will ever need to call this
        method directly. It is called for you when invoking methods on the client.

        This method uses 'cooperative' inheritance due to the nature of external
        constructors. Subclasses should call `super().from_raw` with all of the
        subclass specific parameters passed in `**kwargs`.

        Parameters:
          raw_response (Dict[str, Any]): The raw response from the Kagi API.

          **kwargs: Other parameters to pass to the class constructor.

        Returns:
          KaginawaFastGPTResponse: A new instance of the response object.
        """

        references = [
            KaginawaReference(**ref) for ref in raw_response["data"].get("references", [])
        ]

        raw_response["data"].pop("references", None)

        return super().from_raw(
            raw_response,
            references=references,
            **raw_response["data"],
            **kwargs,
        )


@dataclass
class KaginawaSearchResult:
    t: int
    """Undocumented field of unknown purpose."""

    rank: int
    """The rank of the search result (0 is most relevant)."""

    url: str
    """The URL of the search result."""

    title: str
    """The <title> of the page referenced."""

    snippet: str
    """A short blurb of text describing the result."""

    published: datetime
    """The (self-reported) date the result was published."""

    @classmethod
    def from_raw(cls, raw_result: Dict[str, Any]):
        """Create a model from the raw API response.

        It's not expected that end-users of this library will ever need to call this
        method directly. It is called for you when invoking methods on the client.

        Parameters:
          raw_response (Dict[str, Any]): The raw response from the Kagi API.

        Returns:
          KaginawaSearchResult: A new instance of the result object.
        """

        published = datetime.fromisoformat(raw_result["published"])
        raw_result.pop("published")

        return cls(published=published, **raw_result)


@dataclass
class KaginawaEnrichResponse(KaginawaResponse):
    results: List[KaginawaSearchResult]

    @classmethod
    def from_raw(cls, raw_response: Dict[str, Any], **kwargs):
        """Create a model from the raw API response.

        It's not expected that end-users of this library will ever need to call this
        method directly. It is called for you when invoking methods on the client.

        This method uses 'cooperative' inheritance due to the nature of external
        constructors. Subclasses should call `super().from_raw` with all of the
        subclass specific parameters passed in `**kwargs`.

        Parameters:
          raw_response (Dict[str, Any]): The raw response from the Kagi API.

          **kwargs: Other parameters to pass to the class constructor.

        Returns:
          KaginawaEnrichWebResponse: A new instance of the response object.
        """

        results = [KaginawaSearchResult.from_raw(raw_result) for raw_result in raw_response["data"]]

        return super().from_raw(raw_response, results=results, **kwargs)


@dataclass
class KaginawaSummarizationResponse(KaginawaResponse):
    tokens: int
    """The number of tokens in the response."""

    output: str
    """The summary produced by the API."""

    @classmethod
    def from_raw(cls, raw_response: Dict[str, Any], **kwargs):
        """Create a model from the raw API response.

        It's not expected that end-users of this library will ever need to call this
        method directly. It is called for you when invoking methods on the client.

        This method uses 'cooperative' inheritance due to the nature of external
        constructors. Subclasses should call `super().from_raw` with all of the
        subclass specific parameters passed in `**kwargs`.

        Parameters:
          raw_response (Dict[str, Any]): The raw response from the Kagi API.

          **kwargs: Other parameters to pass to the class constructor.

        Returns:
          KaginawaSummarizationResponse: A new instance of the response object.
        """

        return super().from_raw(
            raw_response,
            tokens=raw_response["data"]["tokens"],
            output=raw_response["data"]["output"],
            **kwargs,
        )


class KaginawaSummarizationEngine(StrEnum):
    """The available summarization engines (as of 2023-11-05)."""

    CECIL = "cecil"
    """Friendly, descriptive, fast summary."""

    AGNES = "agnes"
    """Formal, technical, analytical summary."""

    DAPHNE = "daphne"
    """Informal, creative, friendly summary."""

    MURIEL = "muriel"
    """Best-in-class summary using 'enterprise-grade' model."""


class KaginawaSummaryType(StrEnum):
    """The available summary types (as of 2023-11-05)."""

    SUMMARY = "summary"
    """Paragraph(s) of summary prose."""

    TAKEAWAY = "takeaway"
    """Bulleted list of key points."""



class KaginawaError(Exception):
    """The base exception for the kaginawa library.

    All exceptions will be derived from this class which means you can catch
    all (known) errors produced by this library with the following

        from kaginawa.exceptions import KaginawaError

        try:
            ...
        except KaginawaError:
            ...

    This library (which uses requests under the hood) swallows RequestException
    and reraises it with `raise KaginawaError(...) from RequestException(...)`
    so you don't have to worry about handling those errors directly but they are
    available if needed.
    """

    pass



from datetime import datetime, timedelta

from kaginawa.models import (
    KaginawaEnrichResponse,
    KaginawaFastGPTResponse,
    KaginawaResponse,
    KaginawaSearchResult,
    KaginawaSummarizationResponse,
)


class TestKaginawaModels:
    """Test Kagi dataclasses.

    The only thing that's actually necessary to test is whether they're
    able to parse valid Kagi responses. These tests serve as a reference of
    what format the project *expects* the responses to be so it's easy to
    determine if the code needs updated due to upstream changes."
    """

    def test_kagi_response(self):
        common_kagi_response = {
            "meta": {
                "id": (id := "cbad1ea2-ddef-42e7-af75-81bce2c64924"),
                "node": (node := "us-central1"),
                "ms": (ms := 1),
                "api_balance": (api_balance := 5.0),
            }
        }

        res = KaginawaResponse.from_raw(common_kagi_response)

        assert res.id == id
        assert res.node == node
        assert res.duration == timedelta(milliseconds=ms)
        assert res.api_balance == api_balance

    def test_kagi_fastgpt_response(self):
        fastgpt_kagi_response = {
            "meta": {
                "id": (id := "cbad1ea2-ddef-42e7-af75-81bce2c64924"),
                "node": (node := "us-central1"),
                "ms": (ms := 1),
                "api_balance": (api_balance := 5.0),
            },
            "data": {
                "output": (output := ""),
                "tokens": (tokens := 15),
                "references": (
                    references := [
                        {
                            "title": "Top 10 Cutest Puppies",
                            "snippet": "The definitive list.",
                            "url": "https://example.com/cute-puppies.html",
                        },
                        {
                            "title": "100 Puppers You Don't Want To Miss",
                            "snippet": "You'll wish you were in a pupper puddle.",
                            "url": "https://example.com/recent/puppers",
                        },
                    ]
                ),
            },
        }

        res = KaginawaFastGPTResponse.from_raw(fastgpt_kagi_response)

        assert res.id == id
        assert res.node == node
        assert res.duration == timedelta(milliseconds=ms)
        assert res.api_balance == api_balance

        assert res.output == output
        assert res.tokens == tokens

        assert len(res.references) == len(references)

        for parsed_reference, raw_reference in list(zip(res.references, references)):
            assert parsed_reference.title == raw_reference["title"]
            assert parsed_reference.snippet == raw_reference["snippet"]
            assert parsed_reference.url == raw_reference["url"]

    def test_kagi_fastgpt_response_no_references(self):
        fastgpt_kagi_response = {
            "meta": {
                "id": (id := "cbad1ea2-ddef-42e7-af75-81bce2c64924"),
                "node": (node := "us-central1"),
                "ms": (ms := 1),
                "api_balance": (api_balance := 5.0),
            },
            "data": {
                "output": (output := ""),
                "tokens": (tokens := 15),
            },
        }

        res = KaginawaFastGPTResponse.from_raw(fastgpt_kagi_response)

        assert res.id == id
        assert res.node == node
        assert res.duration == timedelta(milliseconds=ms)
        assert res.api_balance == api_balance

        assert res.output == output
        assert res.tokens == tokens
        assert len(res.references) == 0

    def test_kagi_enrich_response(self):
        enrich_web_response = {
            "meta": {
                "id": (id := "cbad1ea2-ddef-42e7-af75-81bce2c64924"),
                "node": (node := "us-central1"),
                "ms": (ms := 1),
                "api_balance": (api_balance := 5.0),
            },
            "data": (
                results := [
                    {
                        "t": 42,
                        "rank": 0,
                        "url": "https://example.com/which-animal-are-you.html",
                        "title": "Quiz: Which Animal Are You?",
                        "snippet": "Maybe a cat, maybe a dog, find out now!",
                        "published": (datetime.now() - timedelta(days=5)).isoformat(),
                    },
                ]
            ),
        }

        res = KaginawaEnrichResponse.from_raw(enrich_web_response)

        assert res.id == id
        assert res.node == node
        assert res.duration == timedelta(milliseconds=ms)
        assert res.api_balance == api_balance

        assert len(res.results) == len(results)

        for parsed_result, raw_result in list(zip(res.results, results)):
            assert parsed_result.title == raw_result["title"]
            assert parsed_result.snippet == raw_result["snippet"]
            assert parsed_result.url == raw_result["url"]

    def test_kagi_summarization_response(self):
        summarization_kagi_response = {
            "meta": {
                "id": (id := "cbad1ea2-ddef-42e7-af75-81bce2c64924"),
                "node": (node := "us-central1"),
                "ms": (ms := 1),
                "api_balance": (api_balance := 5.0),
            },
            "data": {
                "tokens": (tokens := 42),
                "output": (output := "tl;dr War and Peace: It involves Russia."),
            },
        }

        res = KaginawaSummarizationResponse.from_raw(summarization_kagi_response)

        assert res.id == id
        assert res.node == node
        assert res.duration == timedelta(milliseconds=ms)
        assert res.api_balance == api_balance

        assert res.tokens == tokens
        assert res.output == output

    def test_kagi_search_result(self):
        kagi_search_result = {
            "t": (t := 37),
            "rank": (rank := 0),
            "url": (url := "https://example.com/best-ya-romance-novels-this-year"),
            "title": (title := "Straignt from TikTok: This Year's Best YA Novels"),
            "snippet": (snippet := "Necromancer lesbians and gay love in ancient Greece!"),
            "published": (published := (datetime.now() - timedelta(weeks=7)).isoformat()),
        }

        res = KaginawaSearchResult.from_raw(kagi_search_result)

        assert res.t == t
        assert res.rank == rank
        assert res.title == title
        assert res.url == url
        assert res.snippet == snippet
        assert res.published == datetime.fromisoformat(published)
